#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2016       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Dummy database. This database is always empty and is read only.

It is provided for the initial database on loading Gramps, and also as a
database when a normal database is closed.

Most of the code in Gramps uses dbstate.db.is_open() to determine whether data
can be fetched from a database (essentially to determine whether there is a
database to fetch data from). Thus, dbstate.db cannot be left as 'None' because
None has no 'is_open' attribute. Therefore this database class is provided so
that it can be instantiated for dbstate.db.

FIXME: Ideally, only is_open() needs to be implemented here, because that is the
only method that should really be called, but the Gramps code is not perfect,
and many other methods are called. Calls of other methods could be considered
bugs, so when these are fixed, this class could be reduced.

FIXME: Errors in calling these methods (e.g. accessing data when the database is
closed) should result in exceptions. However, at present (mid-2016) there are so
many cases where these methods are called in error that raising exceptions would
be too disruptive. Hence such errors only result in a warning log message and a
'meaningful' result is returned. When the rest of Gramps code is fixed, these
methods should be changed to generate exceptions. Possibly by globally changing
'LOG.debug' to 'raise DbException'.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations
import inspect
import logging
from abc import ABCMeta
from functools import wraps
from types import FunctionType
from typing import Any

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from ..errors import HandleError
from ..lib import Researcher
from ..utils.callback import Callback
from .base import DbReadBase
from .bookmarks import DbBookmarks
from .dbconst import DBLOGNAME

LOG = logging.getLogger(DBLOGNAME)


# -------------------------------------------------------------------------
#
# some magic, see http://stackoverflow.com/questions/11349183/how-to-wrap-every-
# method-of-a-class
#
# This processes the DummyDb class for diagnostic purposes to wrap each method
# with code to log the method name and where it was called from.
# -------------------------------------------------------------------------
def wrapper(method):
    """
    wrapper method that returns a 'wrapped' method which can be wrapped round
    every function in a class. The 'wrapped' method logs the original function
    that was called, and where it was called from.
    """

    @wraps(method)
    def wrapped(*args, **keywargs):
        """
        This 'wrapped' method logs the original function that was called, and
        where it was called from.
        """
        if __debug__ and LOG.isEnabledFor(logging.DEBUG):
            class_name = args[0].__class__.__name__
            func_name = method.__name__
            frame = inspect.currentframe()
            c_frame = frame.f_back
            c_code = c_frame.f_code
            LOG.debug(
                "calling %s.%s()... from file %s, line %s in %s",
                class_name,
                func_name,
                c_code.co_filename,
                c_frame.f_lineno,
                c_code.co_name,
            )
        return method(*args, **keywargs)

    return wrapped


class MetaClass(type):
    """
    transform class by wrapping it with a diagnostic wrapper (if __debug__ is
    not set)
    """

    def __new__(mcs, class_name, bases, classdict):
        """
        When the class this is applied to (DummyDb) is instantiated, each method
        in the class is wrapped with a diagnostic routine.
        """
        newclassdict = {}
        for attributename, attribute in classdict.items():
            if isinstance(attribute, FunctionType):
                # replace with a wrapped version
                attribute = wrapper(attribute)
            newclassdict[attributename] = attribute
        return type.__new__(mcs, class_name, bases, newclassdict)


class M_A_M_B(ABCMeta, MetaClass):
    """
    Metaclass that inherits from two different metaclasses, so as to avoid
    error: "metaclass conflict: the metaclass of a derived class must be a (non-
    strict) subclass of the metaclasses of all its bases"

    See recipe: http://code.activestate.com/recipes/204197-solving-the-
    metaclass-conflict/
    """


# -------------------------------------------------------------------------
#
# class DummyDb
#
# -------------------------------------------------------------------------
class DummyDb(DbReadBase, Callback, object, metaclass=M_A_M_B):
    """
    Gramps database object. This object is a dummy database class that is always
    empty and is read-only.
    """

    __signals__: dict[str, Any] = {}

    def __init__(self):
        """
        Create a new DummyDb instance.
        """
        Callback.__init__(self)
        self.basedb = None
        self.__feature = {}  # {"feature": VALUE, ...}
        self.db_is_open = False
        self.readonly = True
        self.name_formats = []
        self.bookmarks = DbBookmarks()
        self.family_bookmarks = DbBookmarks()
        self.event_bookmarks = DbBookmarks()
        self.place_bookmarks = DbBookmarks()
        self.citation_bookmarks = DbBookmarks()
        self.source_bookmarks = DbBookmarks()
        self.repo_bookmarks = DbBookmarks()
        self.media_bookmarks = DbBookmarks()
        self.note_bookmarks = DbBookmarks()
        self.owner = Researcher()

    def get_feature(self, feature):
        """
        Databases can implement certain features or not. The default is
        None, unless otherwise explicitly stated.
        """
        return self.__feature.get(feature, None)  # can also be explicitly None

    def set_feature(self, feature, value):
        """
        Databases can implement certain features.
        """
        self.__feature[feature] = value

    def close(self, update=True, user=None):
        """
        Close the specified database.
        """
        if not self.db_is_open:
            LOG.warning("database is already closed")
        else:
            self.db_is_open = False

    def db_has_bm_changes(self):
        """
        Return whether there were bookmark changes during the session.
        """
        if not self.db_is_open:
            LOG.warning("database is already closed")
        LOG.warning("database is readonly")
        return False

    def find_backlink_handles(self, handle, include_classes=None):
        """
        Find all objects that hold a reference to the object handle.

        Returns an iterator over a list of (class_name, handle) tuples.

        :param handle: handle of the object to search for.
        :type handle: database handle
        :param include_classes: list of class names to include in the results.
            Default is None which includes all classes.
        :type include_classes: list of class names

        This default implementation does a sequential scan through all
        the primary object databases and is very slow. Backends can
        override this method to provide much faster implementations that
        make use of additional capabilities of the backend.

        Note that this is a generator function, it returns a iterator for
        use in loops. If you want a list of the results use::

            result_list = list(find_backlink_handles(handle))
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        return []

    def find_initial_person(self):
        """
        Returns first person in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")

    def find_next_event_gramps_id(self):
        """
        Return the next available Gramps ID for a Event object based off the
        event ID prefix.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return ""

    def find_next_family_gramps_id(self):
        """
        Return the next available Gramps ID for a Family object based off the
        family ID prefix.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return ""

    def find_next_note_gramps_id(self):
        """
        Return the next available Gramps ID for a Note object based off the
        note ID prefix.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return ""

    def find_next_media_gramps_id(self):
        """
        Return the next available Gramps ID for a Media object based
        off the media object ID prefix.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return ""

    def find_next_person_gramps_id(self):
        """
        Return the next available Gramps ID for a Person object based off the
        person ID prefix.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return ""

    def find_next_place_gramps_id(self):
        """
        Return the next available Gramps ID for a Place object based off the
        place ID prefix.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return ""

    def find_next_repository_gramps_id(self):
        """
        Return the next available Gramps ID for a Repository object based
        off the repository ID prefix.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return ""

    def find_next_source_gramps_id(self):
        """
        Return the next available Gramps ID for a Source object based off the
        source ID prefix.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return ""

    def get_bookmarks(self):
        """
        Return the list of Person handles in the bookmarks.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return self.bookmarks

    def get_child_reference_types(self):
        """
        Return a list of all child reference types associated with Family
        instances in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_default_handle(self):
        """
        Return the default Person of the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")

    def get_default_person(self):
        """
        Return the default Person of the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")

    def get_event_bookmarks(self):
        """
        Return the list of Event handles in the bookmarks.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return self.event_bookmarks

    def get_event_cursor(self):
        """
        Return a reference to a cursor over event objects
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_event_from_gramps_id(self, gramps_id):
        """
        Find an Event in the database from the passed Gramps ID.

        If no such Event exists, None is returned.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("gramps_id %s does not exist in the dummy database", gramps_id)

    def get_event_from_handle(self, handle):
        """
        Find a Event in the database from the passed Gramps ID.

        If no such Event exists, a HandleError is raised.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)

    def get_event_handles(self):
        """
        Return a list of database handles, one handle for each Event in the
        database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_event_roles(self):
        """
        Return a list of all custom event role names associated with Event
        instances in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_event_attribute_types(self):
        """
        Return a list of all Attribute types assocated with Event instances
        in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_event_types(self):
        """
        Return a list of all event types in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_family_attribute_types(self):
        """
        Return a list of all Attribute types associated with Family instances
        in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_family_bookmarks(self):
        """
        Return the list of Family handles in the bookmarks.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return self.family_bookmarks

    def get_family_cursor(self):
        """
        Return a reference to a cursor over Family objects
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_family_event_types(self):
        """
        Deprecated:  Use get_event_types
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_family_from_gramps_id(self, gramps_id):
        """
        Find a Family in the database from the passed Gramps ID.

        If no such Family exists, None is returned.
        Need to be overridden by the derived class.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("gramps_id %s does not exist in the dummy database", gramps_id)

    def get_family_from_handle(self, handle):
        """
        Find a Family in the database from the passed Gramps ID.

        If no such Family exists, a HandleError is raised.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_family_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Family in
        the database.

        :param sort_handles: If True, the list is sorted by surnames.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_family_relation_types(self):
        """
        Return a list of all relationship types associated with Family
        instances in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_media_attribute_types(self):
        """
        Return a list of all Attribute types associated with Media and MediaRef
        instances in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_media_bookmarks(self):
        """
        Return the list of Media handles in the bookmarks.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return self.media_bookmarks

    def get_media_cursor(self):
        """
        Return a reference to a cursor over Media objects
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_media_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Media in
        the database.

        :param sort_handles: If True, the list is sorted by title.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_mediapath(self):
        """
        Return the default media path of the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return ""

    def get_name_group_keys(self):
        """
        Return the defined names that have been assigned to a default grouping.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_name_group_mapping(self, surname):
        """
        Return the default grouping name for a surname.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return ""

    def get_name_types(self):
        """
        Return a list of all custom names types associated with Person
        instances in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_origin_types(self):
        """
        Return a list of all custom origin types associated with Person/Surname
        instances in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_note_bookmarks(self):
        """
        Return the list of Note handles in the bookmarks.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return self.media_bookmarks

    def get_note_cursor(self):
        """
        Return a reference to a cursor over Note objects
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_note_from_gramps_id(self, gramps_id):
        """
        Find a Note in the database from the passed Gramps ID.

        If no such Note exists, None is returned.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("gramps_id %s does not exist in the dummy database", gramps_id)

    def get_note_from_handle(self, handle):
        """
        Find a Note in the database from the passed Gramps ID.

        If no such Note exists, a HandleError is raised.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_note_handles(self):
        """
        Return a list of database handles, one handle for each Note in the
        database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_note_types(self):
        """
        Return a list of all custom note types associated with Note instances
        in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_number_of_events(self):
        """
        Return the number of events currently in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return 0

    def get_number_of_families(self):
        """
        Return the number of families currently in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return 0

    def get_number_of_media(self):
        """
        Return the number of media objects currently in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return 0

    def get_number_of_notes(self):
        """
        Return the number of notes currently in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return 0

    def get_number_of_people(self):
        """
        Return the number of people currently in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return 0

    def get_number_of_places(self):
        """
        Return the number of places currently in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return 0

    def get_number_of_repositories(self):
        """
        Return the number of source repositories currently in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return 0

    def get_number_of_sources(self):
        """
        Return the number of sources currently in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return 0

    def get_number_of_citations(self):
        """
        Return the number of citations currently in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return 0

    def get_number_of_tags(self):
        """
        Return the number of tags currently in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return 0

    def get_media_from_gramps_id(self, gramps_id):
        """
        Find a Media in the database from the passed Gramps ID.

        If no such Media exists, None is returned.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("gramps_id %s does not exist in the dummy database", gramps_id)

    def get_media_from_handle(self, handle):
        """
        Find an Object in the database from the passed Gramps ID.

        If no such Object exists, a HandleError is raised.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_person_attribute_types(self):
        """
        Return a list of all Attribute types associated with Person instances
        in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_person_cursor(self):
        """
        Return a reference to a cursor over Person objects
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_person_event_types(self):
        """
        Deprecated:  Use get_event_types
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_person_from_gramps_id(self, gramps_id):
        """
        Find a Person in the database from the passed Gramps ID.

        If no such Person exists, None is returned.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("gramps_id %s does not exist in the dummy database", gramps_id)

    def get_person_from_handle(self, handle):
        """
        Find a Person in the database from the passed Gramps ID.

        If no such Person exists, a HandleError is raised.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_person_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Person in
        the database.

        :param sort_handles: If True, the list is sorted by surnames.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_source_attribute_types(self):
        """
        Return a list of all Attribute types associated with Source/Citation
        instances in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_place_bookmarks(self):
        """
        Return the list of Place handles in the bookmarks.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return self.place_bookmarks

    def get_place_cursor(self):
        """
        Return a reference to a cursor over Place objects
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_place_from_gramps_id(self, gramps_id):
        """
        Find a Place in the database from the passed Gramps ID.

        If no such Place exists, None is returned.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("gramps_id %s does not exist in the dummy database", gramps_id)

    def get_place_from_handle(self, handle):
        """
        Find a Place in the database from the passed Gramps ID.

        If no such Place exists, a HandleError is raised.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_place_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Place in
        the database.

        :param sort_handles: If True, the list is sorted by Place title.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_raw_event_data(self, handle):
        """
        Return raw (serialized and pickled) Event object from handle
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_raw_family_data(self, handle):
        """
        Return raw (serialized and pickled) Family object from handle
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_raw_note_data(self, handle):
        """
        Return raw (serialized and pickled) Note object from handle
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_raw_media_data(self, handle):
        """
        Return raw (serialized and pickled) Family object from handle
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_raw_person_data(self, handle):
        """
        Return raw (serialized and pickled) Person object from handle
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_raw_place_data(self, handle):
        """
        Return raw (serialized and pickled) Place object from handle
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_raw_repository_data(self, handle):
        """
        Return raw (serialized and pickled) Repository object from handle
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_raw_source_data(self, handle):
        """
        Return raw (serialized and pickled) Source object from handle
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_raw_citation_data(self, handle):
        """
        Return raw (serialized and pickled) Citation object from handle
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_raw_tag_data(self, handle):
        """
        Return raw (serialized and pickled) Tag object from handle
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_repo_bookmarks(self):
        """
        Return the list of Repository handles in the bookmarks.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return self.repo_bookmarks

    def get_repository_cursor(self):
        """
        Return a reference to a cursor over Repository objects
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_repository_from_gramps_id(self, gramps_id):
        """
        Find a Repository in the database from the passed Gramps ID.

        If no such Repository exists, None is returned.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("gramps_id %s does not exist in the dummy database", gramps_id)

    def get_repository_from_handle(self, handle):
        """
        Find a Repository in the database from the passed Gramps ID.

        If no such Repository exists, a HandleError is raised.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_repository_handles(self):
        """
        Return a list of database handles, one handle for each Repository in
        the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_repository_types(self):
        """
        Return a list of all custom repository types associated with Repository
        instances in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_researcher(self):
        """
        Return the Researcher instance, providing information about the owner
        of the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return self.owner

    def get_save_path(self):
        """
        Return the save path of the file, or "" if one does not exist.
        """
        return ""

    def get_source_bookmarks(self):
        """
        Return the list of Source handles in the bookmarks.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return self.source_bookmarks

    def get_source_cursor(self):
        """
        Return a reference to a cursor over Source objects
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_source_from_gramps_id(self, gramps_id):
        """
        Find a Source in the database from the passed Gramps ID.

        If no such Source exists, None is returned.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("gramps_id %s does not exist in the dummy database", gramps_id)

    def get_source_from_handle(self, handle):
        """
        Find a Source in the database from the passed Gramps ID.

        If no such Source exists, a HandleError is raised.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_source_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Source in
        the database.

        :param sort_handles: If True, the list is sorted by Source title.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_source_media_types(self):
        """
        Return a list of all custom source media types associated with Source
        instances in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_citation_bookmarks(self):
        """
        Return the list of Citation handles in the bookmarks.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return self.citation_bookmarks

    def get_citation_cursor(self):
        """
        Return a reference to a cursor over Citation objects
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_citation_from_gramps_id(self, gramps_id):
        """
        Find a Citation in the database from the passed Gramps ID.

        If no such Citation exists, None is returned.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("gramps_id %s does not exist in the dummy database", gramps_id)

    def get_citation_from_handle(self, handle):
        """
        Find a Citation in the database from the passed Gramps ID.

        If no such Citation exists, a HandleError is raised.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_citation_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Citation in
        the database.

        :param sort_handles: If True, the list is sorted by Citation title.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_surname_list(self):
        """
        Return the list of locale-sorted surnames contained in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_tag_cursor(self):
        """
        Return a reference to a cursor over Tag objects
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_tag_from_handle(self, handle):
        """
        Find a Tag in the database from the passed handle.

        If no such Tag exists, a HandleError is raised.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("handle %s does not exist in the dummy database", handle)
        raise HandleError(f"Handle {handle} not found")

    def get_tag_from_name(self, val):
        """
        Find a Tag in the database from the passed Tag name.

        If no such Tag exists, None is returned.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("tag name %s does not exist in the dummy database", val)

    def get_tag_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Tag in
        the database.

        :param sort_handles: If True, the list is sorted by Tag name.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_url_types(self):
        """
        Return a list of all custom names types associated with Url instances
        in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def get_place_types(self):
        """
        Return a list of all custom place types associated with Place instances
        in the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def has_event_handle(self, handle):
        """
        Return True if the handle exists in the current Event database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return False

    def has_family_handle(self, handle):
        """
        Return True if the handle exists in the current Family database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return False

    def has_name_group_key(self, name):
        """
        Return if a key exists in the name_group table.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return False

    def has_note_handle(self, handle):
        """
        Return True if the handle exists in the current Note database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return False

    def has_media_handle(self, handle):
        """
        Return True if the handle exists in the current Mediadatabase.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return False

    def has_person_handle(self, handle):
        """
        Return True if the handle exists in the current Person database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return False

    def has_place_handle(self, handle):
        """
        Return True if the handle exists in the current Place database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return False

    def has_repository_handle(self, handle):
        """
        Return True if the handle exists in the current Repository database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return False

    def has_source_handle(self, handle):
        """
        Return True if the handle exists in the current Source database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return False

    def has_tag_handle(self, handle):
        """
        Return True if the handle exists in the current Tag database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return False

    def is_open(self):
        """
        Return True if the database has been opened.
        """
        return self.db_is_open

    def iter_citations(self):
        """
        Return an iterator over objects for Citations in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_event_handles(self):
        """
        Return an iterator over handles for Events in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_events(self):
        """
        Return an iterator over objects for Events in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_families(self):
        """
        Return an iterator over objects for Families in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_family_handles(self):
        """
        Return an iterator over handles for Families in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_media_handles(self):
        """
        Return an iterator over handles for Media in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_media(self):
        """
        Return an iterator over objects for Medias in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_note_handles(self):
        """
        Return an iterator over handles for Notes in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_notes(self):
        """
        Return an iterator over objects for Notes in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_people(self):
        """
        Return an iterator over objects for Persons in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_person_handles(self):
        """
        Return an iterator over handles for Persons in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_place_handles(self):
        """
        Return an iterator over handles for Places in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_places(self):
        """
        Return an iterator over objects for Places in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_repositories(self):
        """
        Return an iterator over objects for Repositories in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_repository_handles(self):
        """
        Return an iterator over handles for Repositories in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_source_handles(self):
        """
        Return an iterator over handles for Sources in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_sources(self):
        """
        Return an iterator over objects for Sources in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_tag_handles(self):
        """
        Return an iterator over handles for Tags in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def iter_tags(self):
        """
        Return an iterator over objects for Tags in the database
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return []

    def load(
        self,
        directory,
        callback=None,
        mode=None,
        force_schema_upgrade=False,
        force_bsddb_upgrade=False,
        force_bsddb_downgrade=False,
        force_python_upgrade=False,
        update=True,
    ):
        """
        Open the specified database.
        """
        self.db_is_open = True

    def report_bm_change(self):
        """
        Add 1 to the number of bookmark changes during this session.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("database is readonly")

    def request_rebuild(self):
        """
        Notify clients that the data has changed significantly, and that all
        internal data dependent on the database should be rebuilt.
        Note that all rebuild signals on all objects are emitted at the same
        time. It is correct to assume that this is always the case.

        .. todo:: it might be better to replace these rebuild signals by one
                  single database-rebuild signal.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        LOG.warning("database is readonly")

    def version_supported(self):
        """
        Return True when the file has a supported version.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")
        return True

    def set_event_id_prefix(self, val):
        """
        Set the naming template for Gramps Event ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as E%d or E%04d.
        """
        LOG.warning("database is readonly")

    def set_family_id_prefix(self, val):
        """
        Set the naming template for Gramps Family ID values. The string is
        expected to be in the form of a simple text string, or in a format
        that contains a C/Python style format string using %d, such as F%d
        or F%04d.
        """
        LOG.warning("database is readonly")

    def set_note_id_prefix(self, val):
        """
        Set the naming template for Gramps Note ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as N%d or N%04d.
        """
        LOG.warning("database is readonly")

    def set_media_id_prefix(self, val):
        """
        Set the naming template for Gramps Media ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as O%d or O%04d.
        """
        LOG.warning("database is readonly")

    def set_person_id_prefix(self, val):
        """
        Set the naming template for Gramps Person ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as I%d or I%04d.
        """
        LOG.warning("database is readonly")

    def set_place_id_prefix(self, val):
        """
        Set the naming template for Gramps Place ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as P%d or P%04d.
        """
        LOG.warning("database is readonly")

    def set_prefixes(
        self, person, media, family, source, citation, place, event, repository, note
    ):
        """
        Set the prefixes for the gramps ids for all gramps objects
        """
        LOG.warning("database is readonly")

    def set_repository_id_prefix(self, val):
        """
        Set the naming template for Gramps Repository ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as R%d or R%04d.
        """
        LOG.warning("database is readonly")

    def set_source_id_prefix(self, val):
        """
        Set the naming template for Gramps Source ID values.

        The string is expected to be in the form of a simple text string, or
        in a format that contains a C/Python style format string using %d,
        such as S%d or S%04d.
        """
        LOG.warning("database is readonly")

    def set_mediapath(self, mediapath):
        """
        Set the default media path for database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")

    def set_researcher(self, owner):
        """
        Set the information about the owner of the database.
        """
        if not self.db_is_open:
            LOG.debug("database is closed")

    def get_dbid(self):
        """
        A unique ID for this database on this computer.
        """
        return ""

    def get_dbname(self):
        """
        A name for this database on this computer.
        """
        return ""



# -*- coding: utf-8 -*-

"""Unittest for utils.py"""

import unittest
from ..utils import get_contrast_color


class Test_get_contrast_color(unittest.TestCase):
    def setUp(self):
        self.black = (0, 0, 0)
        self.white = (1, 1, 1)

    def test_contrast_black_returns_white(self):
        contrast_color = get_contrast_color(self.black)
        self.assertEqual(
            contrast_color,
            self.white,
            "Contrasting color for black did not return white",
        )

    def test_contrast_white_returns_black(self):
        contrast_color = get_contrast_color(self.white)
        self.assertEqual(
            contrast_color,
            self.black,
            "Contrasting color for white did not return black",
        )


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2017      Paul Culley
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# ------------------------------------------------------------------------
#
# Python Modules
#
# ------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk


# ------------------------------------------------------------------------
#
# Gramps Modules
#
# ------------------------------------------------------------------------
class CellRendererTextEdit(Gtk.CellRendererText):
    """To be used where you normally use Gtk.CellRendererText and you want to
    avoid losing the text if the user clicks outside the cell (Like an 'OK'
    button."""

    __gtype_name__ = "CellRendererTextEdit"

    def __init__(self):
        Gtk.CellRendererText.__init__(self)

    def do_start_editing(
        self, event, treeview, path, background_area, cell_area, flags
    ):
        if not self.get_property("editable"):
            return
        entry = Gtk.Entry()
        entry.set_has_frame(False)
        xalign, yalign = self.get_alignment()
        entry.set_alignment(xalign)
        entry.set_width_chars(5)
        entry.set_text(self.get_property("text"))  # get original cell text
        entry.add_events(Gdk.EventMask.FOCUS_CHANGE_MASK)
        entry.connect("focus-out-event", self.focus_out, path)
        entry.connect("key-press-event", self._key_press)
        entry.show()
        return entry

    def focus_out(self, entry, event, path):
        self.emit("edited", path, entry.get_text())
        return False

    def _key_press(self, entry, event):
        if event.type == Gdk.EventType.KEY_PRESS:
            if event.keyval == Gdk.KEY_Escape:
                # get original cell text
                entry.set_text(self.get_property("text"))
        return False



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Czech-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Czech parser
#
# -------------------------------------------------------------------------
class DateParserCZ(DateParser):
    """
    Converts a text string into a Date object
    """

    modifier_to_int = {
        "před": Date.MOD_BEFORE,
        "kolem": Date.MOD_ABOUT,
        "po": Date.MOD_AFTER,
        "od": Date.MOD_FROM,
        "do": Date.MOD_TO,
    }

    quality_to_int = {
        "přibližně": Date.QUAL_ESTIMATED,
        "odhadem": Date.QUAL_ESTIMATED,
        "odh.": Date.QUAL_ESTIMATED,
        "vypočteno": Date.QUAL_CALCULATED,
        "vypočtené": Date.QUAL_CALCULATED,
        "vyp.": Date.QUAL_CALCULATED,
    }

    bce = [
        "před naším letopočtem",
        "před Kristem",
        "př. n. l.",
        "př. Kr.",
    ] + DateParser.bce

    def dhformat_changed(self):
        """Allow overriding so a subclass can modify it"""
        # bug 9739 grampslocale.py gets '%-d.%-m.%Y' -- makes it be '%/d.%/m.%Y'
        self.dhformat = self.dhformat.replace("/", "")  # so counteract that

    def init_strings(self):
        DateParser.init_strings(self)
        self._text2 = re.compile(
            r"(\d+)?\.?\s+?%s\.?\s*((\d+)(/\d+)?)?\s*$" % self._mon_str, re.IGNORECASE
        )
        self._span = re.compile(
            r"(od)\s+(?P<start>.+)\s+(do)\s+(?P<stop>.+)", re.IGNORECASE
        )
        self._range = re.compile(
            r"(mezi)\s+(?P<start>.+)\s+(a)\s+(?P<stop>.+)", re.IGNORECASE
        )


# -------------------------------------------------------------------------
#
# Czech display
#
# -------------------------------------------------------------------------
class DateDisplayCZ(DateDisplay):
    """
    Czech language date display class.
    """

    display = DateDisplay.display_formatted

    def formats_changed(self):
        """Allow overriding so a subclass can modify"""

        # bug 9537 grampslocale.py gets '%-d.%-m.%Y' -- makes it be '%/d.%/m.%Y'
        self.dhformat = self.dhformat.replace("/", "")  # so counteract that


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("cs_CZ", "cs", "CS", "Czech", ("%-d.%-m.%Y",)), DateParserCZ, DateDisplayCZ
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010 Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
A module that provides pluggable sidebars.  These provide an interface to
manage pages in the main Gramps window.
"""
# -------------------------------------------------------------------------
#
# GNOME modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.plug import START, END
from .pluginmanager import GuiPluginManager
from .uimanager import ActionGroup

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
UICATEGORY = """      <section id="ViewsInCategory">
        %s
      </section>
    """
UICATEGORYBAR = """    <placeholder id='ViewsInCategoryBar'>
        %s
      </placeholder>
    """

CATEGORY_ICON = {
    "Dashboard": "gramps-gramplet",
    "People": "gramps-person",
    "Relationships": "gramps-relation",
    "Families": "gramps-family",
    "Events": "gramps-event",
    "Ancestry": "gramps-pedigree",
    "Places": "gramps-place",
    "Geography": "gramps-geo",
    "Sources": "gramps-source",
    "Repositories": "gramps-repository",
    "Media": "gramps-media",
    "Notes": "gramps-notes",
    "Citations": "gramps-citation",
}


# -------------------------------------------------------------------------
#
# Navigator class
#
# -------------------------------------------------------------------------
class Navigator:
    """
    A class which defines the graphical representation of the Gramps navigator.
    """

    def __init__(self, viewmanager):
        self.viewmanager = viewmanager
        self.pages = {}
        self._active_page = None
        self.active_cat = None
        self.active_view = None

        self.ui_category = {}
        self.cat_view_group = None
        self.merge_ids = []

        self.top = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.top.show()

        self.select_button = Gtk.ComboBoxText()
        self.select_button.hide()
        self.top.pack_end(self.select_button, False, True, 0)

        self.stack = Gtk.Stack(homogeneous=False)
        self.stack.show()
        self.stack.connect("notify::visible-child-name", self.cb_switch_page)
        self.top.pack_start(self.stack, True, True, 0)

        self.select_button.bind_property(
            "active-id",
            self.stack,
            "visible-child-name",
            GObject.BindingFlags.BIDIRECTIONAL,
        )

    def load_plugins(self, dbstate, uistate):
        """
        Load the sidebar plugins.
        """
        menuitem = """
            <item>
              <attribute name="action">win.ViewInCategory</attribute>
              <attribute name="label" translatable="yes">%s</attribute>
              <attribute name="target">%d %d</attribute>
            </item>
            """
        baritem = """
            <child>
              <object class="GtkToggleToolButton" id="bar%d">
                <property name="action-name">win.ViewInCategory</property>
                <property name="action-target">'%d %d'</property>
                <property name="icon-name">%s</property>
                <property name="tooltip_text" translatable="yes">%s</property>
                <property name="label" translatable="yes">%s</property>
              </object>
              <packing>
                <property name="homogeneous">False</property>
              </packing>
            </child>
            """

        plugman = GuiPluginManager.get_instance()

        categories = []
        views = {}
        for cat_num, cat_views in enumerate(self.viewmanager.get_views()):
            uimenuitems = ""
            uibaritems = ""
            for view_num, page in enumerate(cat_views):
                if view_num == 0:
                    views[cat_num] = []
                    cat_name = page[0].category[1]
                    cat_icon = CATEGORY_ICON.get(page[0].category[0])
                    if cat_icon is None:
                        try:
                            cat_icon = page[0].stock_category_icon
                        except AttributeError:
                            try:
                                cat_icon = page[0].stock_icon
                            except AttributeError:
                                cat_icon = "gramps-view"
                    if cat_icon is None:
                        cat_icon = "gramps-view"
                    categories.append([cat_num, cat_name, cat_icon])

                if view_num < 9:
                    accel = "<PRIMARY><ALT>%d" % ((view_num % 9) + 1)
                    self.viewmanager.uimanager.app.set_accels_for_action(
                        "win.ViewInCategory('%d %d')" % (cat_num, view_num), [accel]
                    )
                uimenuitems += menuitem % (page[0].name, cat_num, view_num)

                stock_icon = page[0].stock_icon
                if stock_icon is None:
                    stock_icon = cat_icon
                uibaritems += baritem % (
                    view_num,
                    cat_num,
                    view_num,
                    stock_icon,
                    page[0].name,
                    page[0].name,
                )

                views[cat_num].append((view_num, page[0].name, stock_icon))

            if len(cat_views) > 1:
                # allow for switching views in a category
                self.ui_category[cat_num] = [
                    UICATEGORY % uimenuitems,
                    UICATEGORYBAR % uibaritems,
                ]

        for pdata in plugman.get_reg_sidebars():
            module = plugman.load_plugin(pdata)
            if not module:
                print("Error loading sidebar '%s': skipping content" % pdata.name)
                continue

            sidebar_class = getattr(module, pdata.sidebarclass)
            sidebar_page = sidebar_class(dbstate, uistate, categories, views)
            self.add(pdata.menu_label, sidebar_page, pdata.order)

    def get_top(self):
        """
        Return the top container widget for the GUI.
        """
        return self.top

    def add(self, title, sidebar, order):
        """
        Add a page to the sidebar for a plugin.
        """
        self.pages[title] = sidebar
        page = sidebar.get_top()
        self.stack.add_named(page, title)

        if order == START:
            self.select_button.prepend(id=title, text=title)
            self.select_button.set_active_id(title)
        else:
            self.select_button.append(id=title, text=title)

        if len(self.stack.get_children()) == 2:
            self.select_button.show_all()

    def view_changed(self, cat_num, view_num):
        """
        Called when a Gramps view is changed.
        """
        self.active_cat = cat_num
        self.active_view = view_num

        # Add buttons to the menu for the different view in the category
        uimanager = self.viewmanager.uimanager
        if self.cat_view_group:
            if self.cat_view_group in uimanager.get_action_groups():
                uimanager.remove_action_group(self.cat_view_group)

            list(map(uimanager.remove_ui, self.merge_ids))

        if cat_num in self.ui_category:
            action = (
                "ViewInCategory",
                self.cb_view_clicked,
                "",
                str(cat_num) + " " + str(view_num),
            )
            self.cat_view_group = ActionGroup("viewmenu", [action])
            uimanager.insert_action_group(self.cat_view_group)
            mergeid = uimanager.add_ui_from_string(self.ui_category[cat_num])
            self.merge_ids.append(mergeid)

        # Call the view_changed method for the active sidebar
        try:
            sidebar = self.pages[self.stack.get_visible_child_name()]
        except KeyError:
            return
        sidebar.view_changed(cat_num, view_num)

    def cb_view_clicked(self, radioaction, value):
        """
        Called when a view is selected from the menu.
        """
        cat_num, view_num = value.get_string().split()
        self.viewmanager.goto_page(int(cat_num), int(view_num))

    def cb_switch_page(self, stack, pspec):
        """
        Called when the user has switched to a new sidebar plugin page.
        """
        old_page = self._active_page
        if old_page is not None:
            self.pages[old_page].inactive()

        title = stack.get_visible_child_name()
        if title is None:
            # May happen during the time that the widgets have been created, but the
            # pages haven't been added.
            return
        self.pages[title].active(self.active_cat, self.active_view)
        if self.active_view is not None:
            self.pages[title].view_changed(self.active_cat, self.active_view)
        self._active_page = title



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2009  B. Malengier
# Copyright (C) 2009  Swoon on bug tracker
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import unittest
import math

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..place import conv_lat_lon


# -------------------------------------------------------------------------
#
# PlaceTest class
#
# -------------------------------------------------------------------------
class PlaceTest(unittest.TestCase):
    def _test_formats_success(self, lat, lon):
        for format_ in ["D.D4", "D.D8", "DEG", "DEG-:", "RT90", "GEDCOM"]:
            res1, res2 = conv_lat_lon(lat, lon, format_)
            self.assertIsNotNone(res1, "Failed for %s %s" % (lat, lon))
            self.assertIsNotNone(res2, "Failed for %s %s" % (lat, lon))

        for format_ in ["ISO-D", "ISO-DM", "ISO-DMS"]:
            res = conv_lat_lon(lat, lon, format_)
            self.assertIsNotNone(res, "Failed for %s %s" % (lat, lon))

    def _test_formats_fail(self, lat, lon):
        res1, res2 = conv_lat_lon(lat, lon)
        self.assertIsNone(res1, "Expected failure for %s %s" % (lat, lon))
        self.assertIsNone(res2, "Expected failure for %s %s" % (lat, lon))

    def test_decimal(self):
        lat, lon = "50.849888888888", "2.885897222222"
        self._test_formats_success(lat, lon)

    def test_dms(self):
        lat, lon = " 50°50'59.60\"N", "  2°53'9.23\"E"
        self._test_formats_success(lat, lon)

    def test_colon(self):
        lat, lon = " 50 : 50 : 59.60 ", " -2:53 : 9.23   "
        self._test_formats_success(lat, lon)

        lat, lon = "+50:1", "-2:1:2"
        self._test_formats_success(lat, lon)

    def test_bad_latitude(self):
        lat, lon = " dummy", "  2#53 ' 9.23  \"  E "
        self._test_formats_fail(lat, lon)

    def test_bad_longitude(self):
        lat, lon = " 50:50: 59.60", "  d u m my"
        self._test_formats_fail(lat, lon)

    def test_dm_ds(self):
        lat, lon = " 50°59.60'N", "  2°53'E"
        self._test_formats_success(lat, lon)

    def test_both_in_latitude(self):
        lat, lon = " 11° 11' 11\" N, 11° 11' 11\" O", " "
        self._test_formats_fail(lat, lon)

    def test_very_small_negative(self):
        lat, lon = "-0.00006", "-0.00006"
        self._test_formats_success(lat, lon)

    def test_missing_direction(self):
        lat, lon = ' 50°59.60"', "  2°53'E"
        self._test_formats_fail(lat, lon)

    def test_wrong_direction(self):
        lat, lon = ' 50°59.60"E', "  2°53'N"
        self._test_formats_fail(lat, lon)

        lat, lon = ' 50°59.99"E', "  2°59'59.99\"N"
        self._test_formats_fail(lat, lon)

    def test_precision(self):
        lat, lon = " 50°59'59.99\"S", "  2°59'59.99\"E"
        self._test_formats_success(lat, lon)

        lat, lon = "N50.849888888888", "E2.885897222222"
        self._test_formats_success(lat, lon)

    def test_large_latitude(self):
        lat, lon = "90.849888888888", "2.885897222222"
        self._test_formats_fail(lat, lon)

        lat, lon = "-91.2", "-1"
        self._test_formats_fail(lat, lon)

    def test_extreme_values(self):
        lat, lon = "90", "-180"
        self._test_formats_success(lat, lon)

        lat, lon = "90° 00' 00.00\" S ", "179° 59'59.99\"W"
        self._test_formats_success(lat, lon)

        lat, lon = "90° 00' 00.00\" N", "180° 00'00.00\" E"
        self._test_formats_fail(lat, lon)

        lat, lon = "90: 00: 00.00 ", "-179: 59:59.99"
        self._test_formats_success(lat, lon)

        lat, lon = "90° 00' 00.00\" N", "180:00:00.00"
        self._test_formats_fail(lat, lon)

        lat, lon = "90", "180"
        self._test_formats_fail(lat, lon)

        lat, lon = " 89°59'60\"N", "  2°53'W"
        self._test_formats_fail(lat, lon)

        lat, lon = " 89°60'00\"N", "  2°53'W"
        self._test_formats_fail(lat, lon)

        lat, lon = " 89.1°40'00\"N", "  2°53'W"
        self._test_formats_fail(lat, lon)

        lat, lon = " 89°40'00\"N", "  2°53.1'W"
        self._test_formats_success(lat, lon)

        lat, lon = "+61° 43' 60.00\"", "+17° 7' 60.00\""
        self._test_formats_fail(lat, lon)

    def test_zero_crossing(self):
        lat, lon = "0", "0"
        self._test_formats_success(lat, lon)

    def test_near_equator(self):
        lat, lon = ' 1°1"N', "  1°1'E"
        self._test_formats_success(lat, lon)

    def test_roundoff(self):
        lat, lon = " 1°59.999'N", "  1°59.999'E"
        self._test_formats_success(lat, lon)

        lat, lon = " 1°59'59.9999\"N", "  1°59'59.9999\"E"
        self._test_formats_success(lat, lon)

        lat, lon = "89°59'59.9999\"S", "179°59'59.9999\"W"
        self._test_formats_success(lat, lon)

        lat, lon = "89°59'59.9999\"N", "179°59'59.9999\"E"
        self._test_formats_success(lat, lon)

    def test_number_of_decimals(self):
        lat, lon = "89°59'59.99999999\"N", "179°59'59.99999999\"E"
        self._test_formats_success(lat, lon)

    def test_quote_notation(self):
        lat, lon = "89°59'59.99'' N", "179°59'59.99''E"
        self._test_formats_success(lat, lon)

    def test_decimal_localization(self):
        lat, lon = "50.849888888888", "2,885897222222"
        self._test_formats_fail(lat, lon)

        lat, lon = "89°59'59.9999\"S", "179°59'59,9999\"W"
        self._test_formats_fail(lat, lon)

        lat, lon = "89°59'1.599,999\"S", "179°59'59,9999\"W"
        self._test_formats_fail(lat, lon)

    def test_large_longitude(self):
        lat, lon = "81.2", "-182.3"
        self._test_formats_fail(lat, lon)

    def test_bad_sign(self):
        lat, lon = "++50:10:1", "2:1:2"
        self._test_formats_fail(lat, lon)

        lat, lon = "-50:10:1", "-+2:1:2"
        self._test_formats_fail(lat, lon)

    def test_missing_minute(self):
        lat, lon = "-50::1", "-2:1:2"
        self._test_formats_fail(lat, lon)

        lat, lon = "+50:", "-2:1:2"
        self._test_formats_fail(lat, lon)

    def test_extra_whitespace(self):
        lat, lon = "- 50  : 2 : 1 ", "-2:1:2"
        self._test_formats_success(lat, lon)

        lat, lon = "+ 50:2 :  1", "-2:1:2"
        self._test_formats_success(lat, lon)

    def test_sign_and_direction(self):
        lat, lon = "+61° 44' 00.00\"N", "+17° 8' 00.00\"E"
        self._test_formats_success(lat, lon)

    def test_extra_colon(self):
        lat, lon = "+50: 0 : 1 : 1", "-2:1:2"
        self._test_formats_fail(lat, lon)

    def test_leading_colon(self):
        lat, lon = ": 0 : 1 : 1", ":1:2"
        self._test_formats_fail(lat, lon)

    def test_degree_symbol(self):
        lat, lon = "N 50º52'21.92\"", "E 124º52'21.92\""
        self._test_formats_success(lat, lon)

        lat, lon = "S 50º52'21.92\"", "W 124º52'21.92\""
        self._test_formats_success(lat, lon)

    def test_RT90_conversion(self):
        """A given lat/lon is converted to RT90 and back"""
        x, y = conv_lat_lon("59:40:9.09", "12:58:57.74", "RT90")
        lat, lon = conv_SWED_RT90_WGS84(float(x), float(y))

        expetced_lat = 59.0 + 40.0 / 60.0 + 9.09 / 3600.0
        expetced_lon = 12.0 + 58.0 / 60.0 + 57.74 / 3600.0

        self.assertAlmostEqual(lat, expetced_lat, places=3)
        self.assertAlmostEqual(lon, expetced_lon, places=3)


def conv_SWED_RT90_WGS84(X, Y):
    """
    Input is X and Y coordinates in RT90 as float
    Output is lat and long in degrees, float as tuple
    """
    # Some constants used for conversion to/from Swedish RT90
    f = 1.0 / 298.257222101
    e2 = f * (2.0 - f)
    n = f / (2.0 - f)
    L0 = math.radians(15.8062845294)  # 15 deg 48 min 22.624306 sec
    k0 = 1.00000561024
    a = 6378137.0  # meter
    at = a / (1.0 + n) * (1.0 + 1.0 / 4.0 * pow(n, 2) + 1.0 / 64.0 * pow(n, 4))
    FN = -667.711  # m
    FE = 1500064.274  # m

    xi = (X - FN) / (k0 * at)
    eta = (Y - FE) / (k0 * at)
    D1 = (
        1.0 / 2.0 * n
        - 2.0 / 3.0 * pow(n, 2)
        + 37.0 / 96.0 * pow(n, 3)
        - 1.0 / 360.0 * pow(n, 4)
    )
    D2 = 1.0 / 48.0 * pow(n, 2) + 1.0 / 15.0 * pow(n, 3) - 437.0 / 1440.0 * pow(n, 4)
    D3 = 17.0 / 480.0 * pow(n, 3) - 37.0 / 840.0 * pow(n, 4)
    D4 = 4397.0 / 161280.0 * pow(n, 4)
    xip = (
        xi
        - D1 * math.sin(2.0 * xi) * math.cosh(2.0 * eta)
        - D2 * math.sin(4.0 * xi) * math.cosh(4.0 * eta)
        - D3 * math.sin(6.0 * xi) * math.cosh(6.0 * eta)
        - D4 * math.sin(8.0 * xi) * math.cosh(8.0 * eta)
    )
    etap = (
        eta
        - D1 * math.cos(2.0 * xi) * math.sinh(2.0 * eta)
        - D2 * math.cos(4.0 * xi) * math.sinh(4.0 * eta)
        - D3 * math.cos(6.0 * xi) * math.sinh(6.0 * eta)
        - D4 * math.cos(8.0 * xi) * math.sinh(8.0 * eta)
    )
    psi = math.asin(math.sin(xip) / math.cosh(etap))
    DL = math.atan2(math.sinh(etap), math.cos(xip))
    LON = L0 + DL
    A = e2 + pow(e2, 2) + pow(e2, 3) + pow(e2, 4)
    B = -1.0 / 6.0 * (7.0 * pow(e2, 2) + 17 * pow(e2, 3) + 30 * pow(e2, 4))
    C = 1.0 / 120.0 * (224.0 * pow(e2, 3) + 889.0 * pow(e2, 4))
    D = 1.0 / 1260.0 * (4279.0 * pow(e2, 4))
    E = (
        A
        + B * pow(math.sin(psi), 2)
        + C * pow(math.sin(psi), 4)
        + D * pow(math.sin(psi), 6)
    )
    LAT = psi + math.sin(psi) * math.cos(psi) * E
    LAT = math.degrees(LAT)
    LON = math.degrees(LON)
    return LAT, LON


if __name__ == "__main__":
    unittest.main()



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008-2012  Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2012-2014  Paul Franklin
# Copyright (C) 2020,2021  Matthias Balk
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
import datetime
import time
from functools import partial
import calendar

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
import gramps.plugins.lib.libholiday as libholiday
from gramps.gen.config import config
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.const import URL_HOMEPAGE
from gramps.gen.datehandler import displayer as date_displayer
from gramps.gen.display.name import displayer as _nd
from gramps.gen.errors import ReportError
from gramps.gen.lib import (
    Date,
    EventRoleType,
    EventType,
    Name,
    NameType,
    Person,
    Surname,
)
from gramps.gen.lib.date import gregorian
from gramps.gen.plug.docgen import (
    FONT_SERIF,
    INDEX_TYPE_TOC,
    PARA_ALIGN_CENTER,
    PARA_ALIGN_LEFT,
    PARA_ALIGN_RIGHT,
    FontStyle,
    GraphicsStyle,
    IndexMark,
    ParagraphStyle,
)
from gramps.gen.plug.docgen.fontscale import string_trim
from gramps.gen.plug.menu import (
    BooleanOption,
    EnumeratedListOption,
    FilterOption,
    NumberOption,
    PersonOption,
    StringOption,
)
from gramps.gen.plug.report import MenuReportOptions, Report, stdoptions, utils
from gramps.gen.utils.alive import probably_alive
from gramps.gen.utils.symbols import Symbols
from gramps.plugins.lib.libholiday import g2iso

_ = glocale.translation.gettext

# ------------------------------------------------------------------------
#
# Constants
#
# ------------------------------------------------------------------------
pt2cm = utils.pt2cm
cm2pt = utils.cm2pt


# _T_ is a gramps-defined keyword -- see po/update_po.py and po/genpot.sh
def _T_(value, context=""):  # enable deferred translations
    return "%s\x04%s" % (context, value) if context else value


_TITLE1 = _T_("My Calendar")
_TITLE2 = _T_("Produced with Gramps")


# ------------------------------------------------------------------------
#
# Calendar
#
# ------------------------------------------------------------------------
class Calendar(Report):
    """
    Create the Calendar object that produces the report.

        incl_private    - Whether to include private data
    """

    def __init__(self, database, options, user):
        Report.__init__(self, database, options, user)
        menu = options.menu
        self._user = user

        stdoptions.run_private_data_option(self, menu)

        get_value = lambda name: menu.get_option_by_name(name).get_value()

        self.year = get_value("year")
        self.name_format = get_value("name_format")
        self.country = get_value("country")
        self.anniversaries = get_value("anniversaries")
        self.start_dow = get_value("start_dow")
        self.maiden_name = get_value("maiden_name")
        self.alive = get_value("alive")
        self.include_death_date = get_value("include_death_date")
        self.birthdays = get_value("birthdays")
        self.text1 = get_value("text1")
        self.text2 = get_value("text2")
        self.text3 = get_value("text3")
        self.filter_option = menu.get_option_by_name("filter")
        self.filter = self.filter_option.get_filter()

        pid = get_value("pid")
        self.center_person = self.database.get_person_from_gramps_id(pid)
        if self.center_person is None:
            raise ReportError(_("Person %s is not in the Database") % pid)

        self.set_locale(get_value("trans"))

        self.__setup_symbols()

    def __setup_symbols(self):
        symbols = Symbols()
        self.symb_bth = symbols.get_symbol_for_string(symbols.SYMBOL_BIRTH)
        self.symb_mrg = symbols.get_symbol_for_string(symbols.SYMBOL_MARRIAGE)
        death_symbol_idx = config.get("utf8.death-symbol")
        self.symb_dth = symbols.get_death_symbol_for_char(death_symbol_idx)

    def get_name(self, person, maiden_name=None):
        """Return person's name, unless maiden_name given,
        unless married_name listed.
        """
        # Get all of a person's names:
        primary_name = person.get_primary_name()
        married_name = None
        names = [primary_name] + person.get_alternate_names()
        for name in names:
            if int(name.get_type()) == NameType.MARRIED:
                married_name = name
                break  # use first
        # Now, decide which to use:
        if maiden_name is not None:
            if married_name is not None:
                name = Name(married_name)
            else:
                name = Name(primary_name)
                surname = Surname()
                surname.set_surname(maiden_name)
                name.set_surname_list([surname])
        else:
            name = Name(primary_name)
        name.set_display_as(self.name_format)
        return _nd.display_name(name)

    def draw_rectangle(self, style, sx, sy, ex, ey):
        """This should be in BaseDoc"""
        self.doc.draw_line(style, sx, sy, sx, ey)
        self.doc.draw_line(style, sx, sy, ex, sy)
        self.doc.draw_line(style, ex, sy, ex, ey)
        self.doc.draw_line(style, sx, ey, ex, ey)

    ### The rest of these all have to deal with calendar specific things

    def add_day_item(self, text, month, day, format="CAL-Text", marks=[None]):
        """Add an item to a day."""
        month_dict = self.calendar.get(month, {})
        day_list = month_dict.get(day, [])
        day_list.append((format, text, marks))
        month_dict[day] = day_list
        self.calendar[month] = month_dict

    def __get_holidays(self):
        """Get the holidays for the specified country and year"""
        holiday_table = libholiday.HolidayTable()
        country = holiday_table.get_countries()[self.country]
        holiday_table.load_holidays(self.year, country)
        for month in range(1, 13):
            for day in range(1, 32):
                holiday_names = holiday_table.get_holidays(month, day)
                for holiday_name in holiday_names:
                    self.add_day_item(self._(holiday_name), month, day, "CAL-Holiday")
                    # FIXME translation only works for a limited set of things
                    # (the right fix is to somehow feed the locale into the
                    # HolidayTable class in plugins/lib/libholiday.py and then
                    # probably changing all the holiday code to somehow defer
                    # the translation of holidays, until it can be based
                    # on the passed-in locale, but since that would probably
                    # also mean checking every use of holidays I don't think
                    # it is advisable to do, with a release so imminent)
                    # it is also debatable whether it is worth bothering at
                    # all, since it is hard for me to imagine why a user would
                    # be wanting to generate a translated report with holidays
                    # since I believe its main use will be for dates of people

    def write_report(self):
        """
        The short method that runs through each month and creates a page.
        """
        # initialize the dict to fill:
        self.calendar = {}

        # get the information, first from holidays:
        if self.country != 0:
            self.__get_holidays()

        # get data from database:
        self.collect_data()
        # generate the report:
        with self._user.progress(
            _("Calendar Report"), _("Formatting months..."), 12
        ) as step:
            for month in range(1, 13):
                step()
                self.print_page(month)

    def print_page(self, month):
        """
        This method actually writes the calendar page.
        """
        style_sheet = self.doc.get_style_sheet()
        ptitle = style_sheet.get_paragraph_style("CAL-Title")
        ptext = style_sheet.get_paragraph_style("CAL-Text")
        pdaynames = style_sheet.get_paragraph_style("CAL-Daynames")
        pnumbers = style_sheet.get_paragraph_style("CAL-Numbers")
        numpos = pt2cm(pnumbers.get_font().get_size())
        ptext1style = style_sheet.get_paragraph_style("CAL-Text1style")
        long_days = self._ldd.long_days

        self.doc.start_page()
        width = self.doc.get_usable_width()
        height = self.doc.get_usable_height()
        header = 2.54  # one inch
        mark = None
        if month == 1:
            mark = IndexMark(self._("Calendar Report"), INDEX_TYPE_TOC, 1)
        self.draw_rectangle("CAL-Border", 0, 0, width, height)
        self.doc.draw_box("CAL-Title", "", 0, 0, width, header, mark)
        self.doc.draw_line("CAL-Border", 0, header, width, header)
        year = self.year
        # assume every calendar header in the world is "<month-name> <year>"
        title = "%s %s" % (
            self._ldd.long_months[month].capitalize(),
            self._get_date(Date(self.year)),
        )  # localized year
        mark = IndexMark(title, INDEX_TYPE_TOC, 2)
        font_height = pt2cm(ptitle.get_font().get_size())
        self.doc.center_text("CAL-Title", title, width / 2, font_height * 0.25, mark)
        cell_width = width / 7
        cell_height = (height - header) / 6
        current_date = datetime.date(year, month, 1)
        spacing = pt2cm(1.25 * ptext.get_font().get_size())  # 158
        if current_date.isoweekday() != g2iso(self.start_dow + 1):
            # Go back to previous first day of week, and start from there
            current_ord = (
                current_date.toordinal()
                - ((current_date.isoweekday() + 7) - g2iso(self.start_dow + 1)) % 7
            )
        else:
            current_ord = current_date.toordinal()
        for day_col in range(7):
            font_height = pt2cm(pdaynames.get_font().get_size())
            self.doc.center_text(
                "CAL-Daynames",
                long_days[(day_col + g2iso(self.start_dow + 1)) % 7 + 1].capitalize(),
                day_col * cell_width + cell_width / 2,
                header - font_height * 1.5,
            )
        for week_row in range(6):
            something_this_week = 0
            for day_col in range(7):
                thisday = current_date.fromordinal(current_ord)
                if thisday.month == month:
                    something_this_week = 1
                    self.draw_rectangle(
                        "CAL-Border",
                        day_col * cell_width,
                        header + week_row * cell_height,
                        (day_col + 1) * cell_width,
                        header + (week_row + 1) * cell_height,
                    )
                    last_edge = (day_col + 1) * cell_width
                    self.doc.center_text(
                        "CAL-Numbers",
                        str(thisday.day),
                        day_col * cell_width + cell_width / 2,
                        header + week_row * cell_height,
                    )
                    list_ = self.calendar.get(month, {}).get(thisday.day, [])
                    # sort the list to get CAL-Holiday on bottom
                    list_.sort(key=lambda x: (x[0], x[1]))
                    position = spacing
                    for format, p, m_list in list_:
                        for line in reversed(p.split("\n")):
                            # make sure text will fit:
                            if position - 0.1 >= cell_height - numpos:  # font daynums
                                break
                            font = ptext.get_font()
                            line = string_trim(font, line, cm2pt(cell_width + 0.2))
                            self.doc.draw_text(
                                format,
                                line,
                                day_col * cell_width + 0.1,
                                header + (week_row + 1) * cell_height - position - 0.1,
                                m_list[0],
                            )
                            if len(m_list) > 1:  # index the spouse too
                                self.doc.draw_text(format, "", 0, 0, m_list[1])
                            position += spacing
                current_ord += 1
        if not something_this_week:
            last_edge = 0
        font_height = pt2cm(1.5 * ptext1style.get_font().get_size())
        x = last_edge + (width - last_edge) / 2
        text1 = str(self.text1)
        if text1 == _(_TITLE1):
            text1 = self._(_TITLE1)
        self.doc.center_text("CAL-Text1style", text1, x, height - font_height * 3)
        text2 = str(self.text2)
        if text2 == _(_TITLE2):
            text2 = self._(_TITLE2)
        self.doc.center_text("CAL-Text2style", text2, x, height - font_height * 2)
        self.doc.center_text("CAL-Text3style", self.text3, x, height - font_height * 1)
        self.doc.end_page()

    def collect_data(self):
        """
        This method runs through the data, and collects the relevant dates
        and text.
        """
        people = self.database.iter_person_handles()
        people = self.filter.apply(self.database, people, user=self._user)

        with self._user.progress(
            _("Calendar Report"), _("Reading database..."), len(people)
        ) as step:
            for person_handle in people:
                step()
                person = self.database.get_person_from_handle(person_handle)

                self._add_birthday(person)
                self._add_anniversaries(person)
                self._add_death_date(person)

    def _add_birthday(self, person):
        if not self.birthdays:
            return

        birth_ref = person.get_birth_ref()
        if not birth_ref:
            return

        def get_surname_of_husband():
            if self.maiden_name not in ["spouse_first", "spouse_last"]:
                return None

            if person.get_gender() == Person.FEMALE:
                family_list = person.get_family_handle_list()
                if family_list:
                    if self.maiden_name == "spouse_first":
                        family_handle = family_list[0]
                    else:
                        family_handle = family_list[-1]
                    family = self.database.get_family_from_handle(family_handle)
                    husband_handle = family.get_father_handle()
                    wife_handle = family.get_mother_handle()
                    if wife_handle == person.get_handle():
                        if husband_handle:
                            father = self.database.get_person_from_handle(
                                husband_handle
                            )
                            if father:
                                return father.get_primary_name().get_surname()

            return None

        birth_event = self.database.get_event_from_handle(birth_ref.ref)
        birth_date = birth_event.get_date_object()

        if birth_date is not None and birth_date.is_valid():
            birth_date = gregorian(birth_date)

            year = birth_date.get_year()
            month = birth_date.get_month()
            day = birth_date.get_day()

            prob_alive_date = Date(self.year, month, day)
            month, day = birth_date.anniversary(self.year)

            nyears = self.year - year
            short_name = self.get_name(person, get_surname_of_husband())
            alive = probably_alive(person, self.database, prob_alive_date)

            if not self.alive or alive:
                if nyears == 0:
                    text = self._("%(person)s, birth") % {"person": short_name}
                else:
                    # to see "nearby" comments
                    ngettext = self._locale.translation.ngettext

                    # Translators: leave all/any {...} untranslated
                    text = ngettext(
                        "{person}, {age}", "{person}, {age}", nyears
                    ).format(person=short_name, age=nyears)

                dth = " (%s)" % (self.symb_dth) if not alive else ""
                text = "%s %s%s" % (self.symb_bth, text, dth)

                self.add_day_item(
                    text,
                    month,
                    day,
                    marks=[utils.get_person_mark(self.database, person)],
                )

    def _add_anniversaries(self, person):
        if not self.anniversaries:
            return

        def is_father(family):
            return person.get_handle() == family.get_father_handle()

        def marriage_still_active(family):
            for event_ref in family.get_event_ref_list():
                if event_ref.get_role() not in [
                    EventRoleType.FAMILY,
                    EventRoleType.PRIMARY,
                ]:
                    continue

                event = self.database.get_event_from_handle(event_ref.ref)
                if event.type in [
                    EventType.DIVORCE,
                    EventType.ANNULMENT,
                    EventType.DIV_FILING,
                ]:
                    return False

            return True

        families = filter(
            marriage_still_active,
            filter(
                is_father,
                map(
                    self.database.get_family_from_handle,
                    person.get_family_handle_list(),
                ),
            ),
        )

        def get_marriage_event(family):
            for event_ref in family.get_event_ref_list():
                if event_ref.get_role() not in [
                    EventRoleType.FAMILY,
                    EventRoleType.PRIMARY,
                ]:
                    continue

                event = self.database.get_event_from_handle(event_ref.ref)
                if event.type in [EventType.MARRIAGE, EventType.MARR_ALT]:
                    return event

            return None

        def get_wife(family):
            mother_handle = family.get_mother_handle()
            if mother_handle:
                return self.database.get_person_from_handle(mother_handle)
            else:
                return None

        short_name = self.get_name(person)

        for family in families:
            wife = get_wife(family)
            event = get_marriage_event(family)
            if event is None or wife is None:
                continue

            event_date = event.get_date_object()

            if event_date.is_valid():
                event_date = gregorian(event_date)

                year = event_date.get_year()
                month = event_date.get_month()
                day = event_date.get_day()

                prob_alive_date = Date(self.year, month, day)

                alive1 = probably_alive(person, self.database, prob_alive_date)
                alive2 = probably_alive(wife, self.database, prob_alive_date)

                nyears = self.year - year
                if nyears == 0:
                    text = self._("%(wife)s and\n %(person)s, wedding") % {
                        "wife": self.get_name(wife),
                        "person": short_name,
                    }
                else:
                    p_dead = " (%s)" % (self.symb_dth) if not alive1 else ""
                    w_dead = " (%s)" % (self.symb_dth) if not alive2 else ""

                    # to see "nearby" comments
                    ngettext = self._locale.translation.ngettext

                    # Translators: leave all/any {...} untranslated
                    text = ngettext(
                        "{wife} and\n {person}, {nyears}",
                        "{wife} and\n {person}, {nyears}",
                        nyears,
                    ).format(
                        wife=self.get_name(wife) + w_dead,
                        person=short_name + p_dead,
                        nyears=nyears,
                    )

                text = "%s %s" % (self.symb_mrg, text)

                if (self.alive and alive1 and alive2) or not self.alive:
                    mark = utils.get_person_mark(self.database, person)
                    wife_mark = utils.get_person_mark(self.database, wife)
                    self.add_day_item(text, month, day, marks=[mark, wife_mark])

    def _add_death_date(self, person):
        if not self.include_death_date:
            return

        death_ref = person.get_death_ref()
        if not death_ref:
            return

        death_event = self.database.get_event_from_handle(death_ref.ref)
        death_date = death_event.get_date_object()

        if death_date is not None and death_date.is_valid():
            death_date = gregorian(death_date)
            text = "{symbol} {name}, {nyears}".format(
                symbol=self.symb_dth,
                name=self.get_name(person),
                nyears=self.year - death_date.get_year(),
            )
            self.add_day_item(
                text,
                death_date.get_month(),
                death_date.get_day(),
                marks=[utils.get_person_mark(self.database, person)],
            )


# ------------------------------------------------------------------------
#
# CalendarOptions
#
# ------------------------------------------------------------------------
class CalendarOptions(MenuReportOptions):
    """Calendar options for graphic calendar"""

    def __init__(self, name, dbase):
        self.__db = dbase
        self.__pid = None
        self.__filter = None
        MenuReportOptions.__init__(self, name, dbase)

    def get_subject(self):
        """Return a string that describes the subject of the report."""
        return self.__filter.get_filter().get_name()

    def add_menu_options(self, menu):
        """Add the options for the graphical calendar"""

        ##########################
        category_name = _("Report Options")
        add_option = partial(menu.add_option, category_name)
        ##########################

        self.__filter = FilterOption(_("Filter"), 0)
        self.__filter.set_help(
            _("Select filter to restrict people that appear on calendar")
        )
        add_option("filter", self.__filter)
        self.__filter.connect("value-changed", self.__filter_changed)

        self.__pid = PersonOption(_("Center Person"))
        self.__pid.set_help(_("The center person for the report"))
        menu.add_option(category_name, "pid", self.__pid)
        self.__pid.connect("value-changed", self.__update_filters)

        text1 = StringOption(_("Text Area 1"), _(_TITLE1))
        text1.set_help(_("First line of text at bottom of calendar"))
        add_option("text1", text1)

        text2 = StringOption(_("Text Area 2"), _(_TITLE2))
        text2.set_help(_("Second line of text at bottom of calendar"))
        add_option("text2", text2)

        text3 = StringOption(_("Text Area 3"), URL_HOMEPAGE)
        text3.set_help(_("Third line of text at bottom of calendar"))
        add_option("text3", text3)

        ##########################
        category_name = _("Report Options (2)")
        add_option = partial(menu.add_option, category_name)
        ##########################

        self._nf = stdoptions.add_name_format_option(menu, category_name)
        self._nf.connect("value-changed", self.__update_filters)

        self.__update_filters()

        stdoptions.add_private_data_option(menu, category_name)

        alive = BooleanOption(_("Include only living people"), True)
        alive.set_help(_("Include only living people in the calendar"))
        add_option("alive", alive)

        stdoptions.add_localization_option(menu, category_name)

        ##########################
        category_name = _("Content")
        add_option = partial(menu.add_option, category_name)
        ##########################

        year = NumberOption(_("Year of calendar"), time.localtime()[0], 1000, 3000)
        year.set_help(_("Year of calendar"))
        add_option("year", year)

        country = EnumeratedListOption(_("Country for holidays"), 0)
        holiday_table = libholiday.HolidayTable()
        countries = holiday_table.get_countries()
        countries.sort()
        if len(countries) == 0 or (len(countries) > 0 and countries[0] != ""):
            countries.insert(0, "")
        count = 0
        for c in countries:
            country.add_item(count, c)
            count += 1
        country.set_help(_("Select the country to see associated holidays"))
        add_option("country", country)

        start_dow = EnumeratedListOption(_("First day of week"), 1)
        long_days = date_displayer.long_days
        for count in range(1, 8):
            # conversion between gramps numbering (sun=1)
            # and iso numbering (mon=1) of weekdays below
            start_dow.add_item((count + 5) % 7 + 1, long_days[count].capitalize())
        start_dow.set_help(_("Select the first day of the week for the calendar"))
        add_option("start_dow", start_dow)

        maiden_name = EnumeratedListOption(_("Birthday surname"), "own")
        maiden_name.add_item(
            "spouse_first", _("Wives use husband's surname (from first family listed)")
        )
        maiden_name.add_item(
            "spouse_last", _("Wives use husband's surname (from last family listed)")
        )
        maiden_name.add_item("own", _("Wives use their own surname"))
        maiden_name.set_help(_("Select married women's displayed surname"))
        add_option("maiden_name", maiden_name)

        birthdays = BooleanOption(_("Include birthdays"), True)
        birthdays.set_help(_("Whether to include birthdays"))
        add_option("birthdays", birthdays)

        anniversaries = BooleanOption(_("Include anniversaries"), True)
        anniversaries.set_help(_("Whether to include anniversaries"))
        add_option("anniversaries", anniversaries)

        include_death_date = BooleanOption(_("Include death dates"), False)
        include_death_date.set_help(_("Include death anniversaries in the calendar"))
        add_option("include_death_date", include_death_date)

    def __update_filters(self):
        """
        Update the filter list based on the selected person
        """
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        nfv = self._nf.get_value()
        filter_list = utils.get_person_filters(
            person, include_single=False, name_format=nfv
        )
        self.__filter.set_filters(filter_list)

    def __filter_changed(self):
        """
        Handle filter change. If the filter is not specific to a person,
        disable the person option
        """
        filter_value = self.__filter.get_value()
        if filter_value == 0:  # "Entire Database" (as "include_single=False")
            self.__pid.set_available(False)
        else:
            # The other filters need a center person (assume custom ones too)
            self.__pid.set_available(True)

    def make_my_style(
        self,
        default_style,
        name,
        description,
        size=9,
        font=FONT_SERIF,
        justified="left",
        color=None,
        align=PARA_ALIGN_CENTER,
        shadow=None,
        italic=0,
        bold=0,
        borders=0,
        indent=None,
    ):
        """Create paragraph and graphic styles of the same name"""
        # Paragraph:
        f = FontStyle()
        f.set_size(size)
        f.set_type_face(font)
        f.set_italic(italic)
        f.set_bold(bold)
        p = ParagraphStyle()
        p.set_font(f)
        p.set_alignment(align)
        p.set_description(description)
        p.set_top_border(borders)
        p.set_left_border(borders)
        p.set_bottom_border(borders)
        p.set_right_border(borders)
        if indent:
            p.set(first_indent=indent)
        if justified == "left":
            p.set_alignment(PARA_ALIGN_LEFT)
        elif justified == "right":
            p.set_alignment(PARA_ALIGN_RIGHT)
        elif justified == "center":
            p.set_alignment(PARA_ALIGN_CENTER)
        default_style.add_paragraph_style(name, p)
        # Graphics:
        g = GraphicsStyle()
        g.set_paragraph_style(name)
        if shadow:
            g.set_shadow(*shadow)
        if color is not None:
            g.set_fill_color(color)
        if not borders:
            g.set_line_width(0)
        default_style.add_draw_style(name, g)

    def make_default_style(self, default_style):
        """Add the styles used in this report"""
        self.make_my_style(
            default_style,
            "CAL-Title",
            _("Title text and background color"),
            20,
            bold=1,
            italic=1,
            color=(0xEA, 0xEA, 0xEA),
        )
        self.make_my_style(
            default_style, "CAL-Numbers", _("Calendar day numbers"), 13, bold=1
        )
        self.make_my_style(default_style, "CAL-Text", _("Daily text display"), 9)
        self.make_my_style(
            default_style, "CAL-Holiday", _("Holiday text display"), 9, bold=1, italic=1
        )
        self.make_my_style(
            default_style,
            "CAL-Daynames",
            _("Days of the week text"),
            12,
            italic=1,
            bold=1,
            color=(0xEA, 0xEA, 0xEA),
        )
        self.make_my_style(
            default_style, "CAL-Text1style", _("Text at bottom, line 1"), 12
        )
        self.make_my_style(
            default_style, "CAL-Text2style", _("Text at bottom, line 2"), 12
        )
        self.make_my_style(
            default_style, "CAL-Text3style", _("Text at bottom, line 3"), 9
        )
        self.make_my_style(default_style, "CAL-Border", _("Borders"), borders=True)



from gramps.gen.db.utils import open_database
from gramps.gen.db.conversion_tools import convert_21

# 1. Prepare: create a database named "Example" in version 20
#    containing the blob data

# 2. Open the database in latest Gramps to convert the database
#    to version 21

db = open_database("Example")

# This is a version 21 database

# But we tell it to use the blob data:

db.set_serializer("blob")

for table_name in db._get_table_func():
    print("Testing %s..." % table_name)
    get_array_from_handle = db._get_table_func(table_name, "raw_func")
    iter_objects = db._get_table_func(table_name, "iter_func")

    for obj in iter_objects():
        # We convert the object into the JSON dicts:
        json_data = db.serializer.object_to_data(obj)

        # We get the blob array:
        array = get_array_from_handle(obj.handle)

        # We convert the array to JSON dict using the
        # conversion code to convert array directly to
        # dict

        convert_data = convert_21(table_name, array)

        # Now we make sure they are identical in types
        # and values:
        assert convert_data == json_data

db.close()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# gui/editors/__init__.py

from .editaddress import EditAddress
from .editattribute import EditAttribute, EditSrcAttribute
from .editchildref import EditChildRef
from .editcitation import EditCitation
from .editdate import EditDate
from .editevent import EditEvent
from .editeventref import EditEventRef
from .editfamily import EditFamily
from .editldsord import EditLdsOrd, EditFamilyLdsOrd
from .editlocation import EditLocation
from .editmedia import EditMedia
from .editmediaref import EditMediaRef
from .editname import EditName
from .editnote import EditNote
from .editperson import EditPerson
from .editpersonref import EditPersonRef
from .editplace import EditPlace
from .editplacename import EditPlaceName
from .editplaceref import EditPlaceRef
from .editrepository import EditRepository
from .editreporef import EditRepoRef
from .editsource import EditSource
from .edittaglist import EditTagList
from .editurl import EditUrl
from .editlink import EditLink
from .filtereditor import FilterEditor, EditFilter
from gramps.gen.lib import (
    Person,
    Family,
    Event,
    Place,
    Repository,
    Source,
    Citation,
    Media,
    Note,
)

# Map from gramps.gen.lib name to Editor:
EDITORS = {
    "Person": EditPerson,
    "Event": EditEvent,
    "Family": EditFamily,
    "Media": EditMedia,
    "Source": EditSource,
    "Citation": EditCitation,
    "Place": EditPlace,
    "Repository": EditRepository,
    "Note": EditNote,
}

CLASSES = {
    "Person": Person,
    "Event": Event,
    "Family": Family,
    "Media": Media,
    "Source": Source,
    "Citation": Citation,
    "Place": Place,
    "Repository": Repository,
    "Note": Note,
}


def EditObject(
    dbstate, uistate, track, obj_class, prop=None, value=None, callback=None
):
    """
    Generic Object Editor.
    obj_class is Person, Source, Repository, etc.
    prop is 'handle', 'gramps_id', or None (for new object)
    value is string handle, string gramps_id, or None (for new object)
    """
    import logging

    LOG = logging.getLogger(".Edit")
    if obj_class in EDITORS.keys():
        if value is None:
            obj = CLASSES[obj_class]()
            try:
                EDITORS[obj_class](dbstate, uistate, track, obj, callback=callback)
            except Exception as msg:
                LOG.warning(str(msg))
        elif prop in ("gramps_id", "handle"):
            obj = dbstate.db.method("get_%s_from_%s", obj_class, prop)(value)
            if obj:
                try:
                    EDITORS[obj_class](dbstate, uistate, track, obj, callback=callback)
                except Exception as msg:
                    LOG.warning(str(msg))
            else:
                LOG.warning("gramps://%s/%s/%s not found" % (obj_class, prop, value))
        else:
            LOG.warning(
                "unknown property to edit '%s'; "
                "should be 'gramps_id' or 'handle'" % prop
            )
    else:
        LOG.warning(
            "unknown object to edit '%s'; "
            "should be one of %s" % (obj_class, list(EDITORS.keys()))
        )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2012       Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Database utilites
"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
import logging
import os

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from ..config import config
from ..const import GRAMPS_LOCALE as glocale
from ..const import PLUGINS_DIR, USER_PLUGINS
from ..constfunc import get_env_var, win
from ..lib import NameOriginType
from ..plug import BasePluginManager
from .dbconst import DBBACKEND, DBLOCKFN, DBLOGNAME

_ = glocale.translation.gettext

_LOG = logging.getLogger(DBLOGNAME)


def make_database(plugin_id):
    """
    Make a database, given a plugin id.
    """
    pmgr = BasePluginManager.get_instance()
    pdata = pmgr.get_plugin(plugin_id)

    if not pdata:
        # This might happen if using gramps from outside, and
        # we haven't loaded plugins yet
        pmgr.reg_plugins(PLUGINS_DIR, None, None)
        pmgr.reg_plugins(USER_PLUGINS, None, None, load_on_reg=True)
        pdata = pmgr.get_plugin(plugin_id)

    if pdata:
        mod = pmgr.load_plugin(pdata)
        if mod:
            database = getattr(mod, pdata.databaseclass)
            db = database()
            if __debug__ and _LOG.isEnabledFor(logging.DEBUG):
                import inspect

                frame = inspect.currentframe()
                c_frame = frame.f_back
                c_code = c_frame.f_code
                _LOG.debug(
                    "Database class instance created Class:%s instance:%s. "
                    "Called from File %s, line %s, in %s",
                    db.__class__.__name__,
                    hex(id(db)),
                    c_code.co_filename,
                    c_frame.f_lineno,
                    c_code.co_name,
                )
            return db
        raise Exception(f"can't load database backend: '{plugin_id}'")
    raise Exception(f"no such database backend: '{plugin_id}'")


def open_database(dbname, force_unlock=False, callback=None):
    """
    Open a database by name and return the database.
    """
    data = lookup_family_tree(dbname)
    database = None
    if data:
        dbpath, locked, _, backend = data
        if (not locked) or (locked and force_unlock):
            database = make_database(backend)
            database.load(dbpath, callback=callback)
    return database


def lookup_family_tree(dbname):
    """
    Find a Family Tree given its name, and return properties.
    """
    dbdir = os.path.expanduser(config.get("database.path"))
    for dpath in os.listdir(dbdir):
        dirpath = os.path.join(dbdir, dpath)
        path_name = os.path.join(dirpath, "name.txt")
        if os.path.isfile(path_name):
            with open(path_name, "r", encoding="utf8") as file:
                name = file.readline().strip()
            if dbname == name:
                locked = False
                locked_by = None
                backend = get_dbid_from_path(dirpath)
                try:
                    fname = os.path.join(dirpath, "lock")
                    with open(fname, "r", encoding="utf8") as ifile:
                        locked_by = ifile.read().strip()
                        locked = True
                except (OSError, IOError):
                    pass
                return (dirpath, locked, locked_by, backend)
    return None


def get_dbid_from_path(dirpath):
    """
    Return a database backend from a directory path.
    """
    dbid = "bsddb"
    dbid_path = os.path.join(dirpath, DBBACKEND)
    if os.path.isfile(dbid_path):
        with open(dbid_path, encoding="utf8") as file:
            dbid = file.read().strip()
    return dbid


def import_as_dict(filename, user, skp_imp_adds=True):
    """
    Import the filename into a InMemoryDB and return it.
    """
    db = make_database("sqlite")
    db.load(":memory:")
    db.set_feature("skip-import-additions", skp_imp_adds)
    db.set_prefixes(
        config.get("preferences.iprefix"),
        config.get("preferences.oprefix"),
        config.get("preferences.fprefix"),
        config.get("preferences.sprefix"),
        config.get("preferences.cprefix"),
        config.get("preferences.pprefix"),
        config.get("preferences.eprefix"),
        config.get("preferences.rprefix"),
        config.get("preferences.nprefix"),
    )
    status = import_from_filename(db, filename, user)
    return db if status else None


def import_from_filename(db, filename, user):
    """
    Import a file into a database.
    """
    (_, ext) = os.path.splitext(os.path.basename(filename))
    extension = ext[1:].lower()
    pmgr = BasePluginManager.get_instance()
    import_list = pmgr.get_reg_importers()
    for pdata in import_list:
        if extension == pdata.extension:
            mod = pmgr.load_plugin(pdata)
            if not mod:
                return False
            import_function = getattr(mod, pdata.import_function)
            results = import_function(db, filename, user)
            if results is None:
                return False
            return True
    return False


def find_surname_name(surname_list):
    """
    Create a surname-string to use for sort and index.

    All non pa/matronymic surnames are used in indexing.
    pa/matronymic not as they change for every generation!

    Returns a string.
    """
    if surname_list:
        surn = " ".join(
            [
                surname.surname
                for surname in surname_list
                if surname.origintype.value
                not in [NameOriginType.PATRONYMIC, NameOriginType.MATRONYMIC]
            ]
        )
    else:
        surn = ""
    return surn


def clear_lock_file(name):
    """
    Clear the lock file.
    """
    try:
        os.unlink(os.path.join(name, DBLOCKFN))
    except OSError:
        return


def write_lock_file(name):
    """
    Write the lock file.
    """
    if not os.path.isdir(name):
        os.mkdir(name)
    with open(os.path.join(name, DBLOCKFN), "w", encoding="utf8") as lock_file:
        if win():
            user = get_env_var("USERNAME")
            host = get_env_var("USERDOMAIN")
            if not user:
                user = _("Unknown")
        else:
            host = os.uname()[1]
            # An ugly workaround for os.getlogin() issue with Konsole
            try:
                user = os.getlogin()
            except:
                # not win, so don't need get_env_var.
                # under cron getlogin() throws and there is no USER.
                user = os.environ.get("USER", "noUSER")
        if host:
            text = f"{user}@{host}"
        else:
            text = user
        # Save only the username and host, so the massage can be
        # printed with correct locale in DbManager.py when a lock is found
        lock_file.write(text)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2008       B. Malengier
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2012       Nick Hall
# Copyright (C) 2024       Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Written by Alex Roitman

"""Tools/Utilities/Media Manager"""

# ------------------------------------------------------------------------
#
# standard python modules
#
# ------------------------------------------------------------------------
import os

# ------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# ------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject
from gi.repository import GdkPixbuf

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import URL_MANUAL_PAGE, ICON, SPLASH
from gramps.gui.display import display_help
from gramps.gen.lib import Media

from gramps.gen.lib.json_utils import data_to_object
from gramps.gen.db import DbTxn
from gramps.gen.updatecallback import UpdateCallback
from gramps.gui.plug import tool
from gramps.gen.utils.file import media_path_full, relative_path, media_path
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.mime import get_type, is_image_type
from gramps.gui.managedwindow import ManagedWindow

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = "%s_-_Tools" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Media_Manager", "manual")


# -------------------------------------------------------------------------
#
# This is an Assistant implementation to guide the user
#
# -------------------------------------------------------------------------
class MediaMan(ManagedWindow, tool.Tool):
    def __init__(self, dbstate, user, options_class, name, callback=None):
        uistate = user.uistate

        tool.Tool.__init__(self, dbstate, options_class, name)
        ManagedWindow.__init__(self, uistate, [], self.__class__)
        self.callback = uistate.pulse_progressbar

        self.batch_ops = []
        self.build_batch_ops()

        self.assistant = Gtk.Assistant()
        self.set_window(self.assistant, None, _("Media Manager"))
        self.setup_configs("interface.mediaman", 780, 600)

        help_btn = Gtk.Button.new_with_label(_("Help"))
        help_btn.connect("clicked", self.on_help_clicked)
        self.assistant.add_action_widget(help_btn)

        self.assistant.connect("close", self.do_close)
        self.assistant.connect("cancel", self.do_close)
        self.assistant.connect("apply", self.run)
        self.assistant.connect("prepare", self.prepare)

        intro = IntroductionPage()
        self.add_page(intro, Gtk.AssistantPageType.INTRO, _("Introduction"))
        self.selection = SelectionPage(self.batch_ops)
        self.add_page(self.selection, Gtk.AssistantPageType.CONTENT, _("Selection"))
        self.settings = SettingsPage(self.batch_ops, self.assistant)
        self.add_page(self.settings, Gtk.AssistantPageType.CONTENT)
        self.confirmation = ConfirmationPage(self.batch_ops)
        self.add_page(
            self.confirmation, Gtk.AssistantPageType.CONFIRM, _("Final confirmation")
        )
        self.conclusion = ConclusionPage(self.assistant)
        self.add_page(self.conclusion, Gtk.AssistantPageType.SUMMARY)

        self.show()
        self.assistant.set_forward_page_func(self.forward_page, None)

    def build_menu_names(self, obj):
        """Override :class:`.ManagedWindow` method."""
        return (_("Media Manager"), None)

    def do_close(self, assistant):
        """
        Close the assistant.
        """
        position = self.window.get_position()  # crock
        self.assistant.hide()
        self.window.move(position[0], position[1])
        self.close()

    def forward_page(self, page, data):
        """
        Specify the next page to be displayed.
        """
        if page == 1:  # selection page
            index = self.selection.get_index()
            if self.settings.prepare(index):
                return page + 1
            else:
                return page + 2
        else:
            return page + 1

    def prepare(self, assistant, page):
        """
        Run page preparation code.
        """
        if self.assistant.get_current_page() == 3:
            index = self.selection.get_index()
            self.confirmation.prepare(index)
        self.assistant.set_page_complete(page, True)

    def add_page(self, page, page_type, title=""):
        """
        Add a page to the assistant.
        """
        page.show_all()
        self.assistant.append_page(page)
        self.assistant.set_page_title(page, title)
        self.assistant.set_page_type(page, page_type)

    def on_help_clicked(self, obj):
        """
        Display the relevant portion of Gramps manual.
        """
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def build_batch_ops(self):
        """
        Define the batch operations available.
        """
        batches_to_use = [
            PathChange,
            Convert2Abs,
            Convert2Rel,
            ImagesNotIncluded,
        ]

        for batch_class in batches_to_use:
            self.batch_ops.append(batch_class(self.db, self.callback))

    def run(self, assistant):
        """
        Run selected batch op with selected settings.
        """
        index = self.selection.get_index()
        self.pre_run()
        success = self.batch_ops[index].run_tool()
        self.conclusion.set_result(success)
        self.post_run()

    def pre_run(self):
        """
        Code to run prior to the batch op.
        """
        self.uistate.set_busy_cursor(True)
        self.uistate.progress.show()

    def post_run(self):
        """
        Code to run after to the batch op.
        """
        self.uistate.set_busy_cursor(False)
        self.uistate.progress.hide()


# ------------------------------------------------------------------------
#
# Assistant pages
#
# ------------------------------------------------------------------------
class IntroductionPage(Gtk.Box):
    """
    A page containing introductory text.
    """

    def __init__(self):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        # Using set_page_side_image causes window sizing problems, so put the
        # image in the main page instead.
        image = Gtk.Image()
        image.set_from_file(SPLASH)

        label = Gtk.Label(label=self.__get_intro_text())
        label.set_line_wrap(True)
        label.set_use_markup(True)
        label.set_max_width_chars(60)

        self.pack_start(image, False, False, 0)
        self.pack_start(label, False, False, 5)

    def __get_intro_text(self):
        """
        Return the introductory text.
        """
        return _(
            "This tool allows batch operations on media objects "
            "stored in Gramps. "
            "An important distinction must be made between a Gramps "
            "media object and its file.\n\n"
            "The Gramps media object is a collection of data about "
            "the media object file: its filename and/or path, its "
            "description, its ID, notes, source references, etc. "
            "These data "
            "%(bold_start)sdo not include the file itself%(bold_end)s.\n\n"
            "The files containing image, sound, video, etc, exist "
            "separately on your hard drive. These files are "
            "not managed by Gramps and are not included in the Gramps "
            "database. "
            "The Gramps database only stores the path and file names.\n\n"
            "This tool allows you to only modify the records within "
            "your Gramps database. If you want to move or rename "
            "the files then you need to do it on your own, outside of "
            "Gramps. Then you can adjust the paths using this tool so "
            "that the media objects store the correct file locations."
        ) % {"bold_start": "<b>", "bold_end": "</b>"}


class SelectionPage(Gtk.Box):
    """
    A page with the radio buttons for every available batch op.
    """

    def __init__(self, batch_ops):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        self.batch_op_buttons = []

        self.set_spacing(12)

        grid = Gtk.Grid()
        grid.set_row_spacing(6)
        grid.set_column_spacing(6)

        button = None
        for index in range(len(batch_ops)):
            title = batch_ops[index].title
            description = batch_ops[index].description

            button = Gtk.RadioButton.new_with_mnemonic_from_widget(button, title)
            button.set_tooltip_text(description)
            self.batch_op_buttons.append(button)
            grid.attach(button, 0, 2 * index, 2, 1)

        self.add(grid)

    def get_index(self):
        """
        Query the selection radiobuttons and return the index number
        of the selected batch op.
        """
        for index in range(len(self.batch_op_buttons)):
            button = self.batch_op_buttons[index]
            if button.get_active():
                return index

        return 0


class SettingsPage(Gtk.Box):
    """
    An extra page with the settings specific for the chosen batch-op.
    """

    def __init__(self, batch_ops, assistant):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        self.assistant = assistant
        self.batch_ops = batch_ops

    def prepare(self, index):
        """
        Build the settings for the batch op.
        """
        config = self.batch_ops[index].build_config()
        if config:
            title, contents = config
            self.assistant.set_page_title(self, title)
            list(map(self.remove, self.get_children()))
            self.pack_start(contents, True, True, 0)
            self.show_all()
            return True
        else:
            self.assistant.set_page_title(self, "")
            return False


class ConfirmationPage(Gtk.Box):
    """
    A page to display the summary of the proposed action, as well as the
    list of affected paths.
    """

    def __init__(self, batch_ops):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        self.batch_ops = batch_ops

        self.set_spacing(12)
        self.set_border_width(12)

        self.confirm = Gtk.Label()
        self.confirm.set_halign(Gtk.Align.START)
        self.confirm.set_line_wrap(True)
        self.confirm.set_use_markup(True)
        self.pack_start(self.confirm, False, True, 0)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.set_shadow_type(Gtk.ShadowType.IN)
        tree = Gtk.TreeView()
        self.path_model = Gtk.ListStore(GObject.TYPE_STRING)
        tree.set_model(self.path_model)
        tree_view_column = Gtk.TreeViewColumn(
            _("Affected path"), Gtk.CellRendererText(), text=0
        )
        tree_view_column.set_sort_column_id(0)
        tree.append_column(tree_view_column)
        scrolled_window.add(tree)
        self.pack_start(scrolled_window, True, True, 0)

        label3 = Gtk.Label(
            label=_(
                "Press Apply to proceed, Cancel to abort, "
                "or Back to revisit your options."
            )
        )
        self.pack_start(label3, False, True, 0)

    def prepare(self, index):
        """
        Display a list of changes to be made.
        """
        confirm_text = self.batch_ops[index].build_confirm_text()
        path_list = self.batch_ops[index].build_path_list()

        self.confirm.set_text(confirm_text)

        self.path_model.clear()
        for path in path_list:
            self.path_model.append(row=[path])


class ConclusionPage(Gtk.Box):
    """
    A page to display the summary of the proposed action, as well as the
    list of affected paths.
    """

    def __init__(self, assistant):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        self.assistant = assistant

        image = Gtk.Image()
        image.set_from_file(SPLASH)

        self.label = Gtk.Label()
        self.label.set_line_wrap(True)

        self.pack_start(image, False, False, 0)
        self.pack_start(self.label, False, False, 5)

    def set_result(self, success):
        if success:
            conclusion_title = _("Operation successfully finished")
            conclusion_text = _(
                "The operation you requested has finished successfully. "
                "You may press Close now to continue."
            )
        else:
            conclusion_title = _("Operation failed")
            conclusion_text = _(
                "There was an error while performing the requested "
                "operation. You may try starting the tool again."
            )
        self.label.set_text(conclusion_text)
        self.assistant.set_page_title(self, conclusion_title)


# ------------------------------------------------------------------------
#
# These are the actuall sub-tools (batch-ops) for use from Assistant
#
# ------------------------------------------------------------------------
class BatchOp(UpdateCallback):
    """
    Base class for the sub-tools.
    """

    title = "Untitled operation"
    description = "This operation needs to be described"

    def __init__(self, db, callback):
        UpdateCallback.__init__(self, callback)
        self.db = db
        self.prepared = False

    def build_config(self):
        """
        This method should return either None (if the batch op requires
        no settings to run) or a tuple (title,box) for the settings page.
        """
        return None

    def build_confirm_text(self):
        """
        This method should return either None (if the batch op requires
        no confirmation) or a string with the confirmation text.
        """
        text = _(
            "The following action is to be performed:\n\n" "Operation:\t%s"
        ) % self.title.replace("_", "")
        return text

    def build_path_list(self):
        """
        This method returns a list of the path names that would be
        affected by the batch op. Typically it would rely on prepare()
        to do the actual job, but it does not have to be that way.
        """
        self.prepare()
        return self.path_list

    def run_tool(self):
        """
        This method runs the batch op, taking care of database signals
        and transactions before and after the running.
        Should not be overridden without good reasons.
        """
        self.db.disable_signals()
        with DbTxn(self.title, self.db, batch=True) as self.trans:
            success = self._run()
        self.db.enable_signals()
        self.db.request_rebuild()
        return success

    def _run(self):
        """
        This method is the beef of the tool.
        Needs to be overridden in the subclass.
        """
        print("This method needs to be written.")
        print("Running BatchOp tool... done.")
        return True

    def prepare(self):
        """
        This method should prepare the tool for the actual run.
        Typically this involves going over media objects and
        selecting the ones that will be affected by the batch op.

        This method should set self.prepared to True, to indicate
        that it has already ran.
        """
        self.handle_list = []
        self.path_list = []
        self._prepare()
        self.prepared = True

    def _prepare(self):
        print("This method needs to be written.")
        print("Preparing BatchOp tool... done.")


# ------------------------------------------------------------------------
# Simple op to replace substrings in the paths
# ------------------------------------------------------------------------
class PathChange(BatchOp):
    title = _("Replace _substrings in the path")
    description = _(
        "This tool allows replacing specified substring in the "
        "path of media objects with another substring. "
        "This can be useful when you move your media files "
        "from one directory to another"
    )

    def build_config(self):
        title = _("Replace substring settings")

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.set_spacing(12)

        grid = Gtk.Grid()
        grid.set_row_spacing(6)
        grid.set_column_spacing(6)

        self.from_entry = Gtk.Entry()
        self.from_entry.set_hexpand(True)
        grid.attach(self.from_entry, 1, 0, 1, 1)

        from_label = Gtk.Label(label=_("_Replace:"))
        from_label.set_halign(Gtk.Align.START)
        from_label.set_use_underline(True)
        from_label.set_mnemonic_widget(self.from_entry)
        grid.attach(from_label, 0, 0, 1, 1)

        self.to_entry = Gtk.Entry()
        self.to_entry.set_hexpand(True)
        grid.attach(self.to_entry, 1, 1, 1, 1)

        to_label = Gtk.Label(label=_("_With:"))
        to_label.set_halign(Gtk.Align.START)
        to_label.set_use_underline(True)
        to_label.set_mnemonic_widget(self.to_entry)
        grid.attach(to_label, 0, 1, 1, 1)

        box.add(grid)

        return (title, box)

    def build_confirm_text(self):
        from_text = str(self.from_entry.get_text())
        to_text = str(self.to_entry.get_text())
        text = _(
            "The following action is to be performed:\n\n"
            "Operation:\t%(title)s\n"
            "Replace:\t\t%(src_fname)s\n"
            "With:\t\t%(dest_fname)s"
        ) % {
            "title": self.title.replace("_", ""),
            "src_fname": from_text,
            "dest_fname": to_text,
        }
        return text

    def _prepare(self):
        from_text = str(self.from_entry.get_text())
        self.set_total(self.db.get_number_of_media())
        with self.db.get_media_cursor() as cursor:
            for handle, data in cursor:
                obj = data_to_object(data)
                if obj.get_path().find(from_text) != -1:
                    self.handle_list.append(handle)
                    self.path_list.append(obj.path)
                self.update()
        self.reset()
        self.prepared = True

    def _run(self):
        if not self.prepared:
            self.prepare()
        self.set_total(len(self.handle_list))
        from_text = str(self.from_entry.get_text())
        to_text = str(self.to_entry.get_text())
        for handle in self.handle_list:
            obj = self.db.get_media_from_handle(handle)
            new_path = obj.get_path().replace(from_text, to_text)
            obj.set_path(new_path)
            self.db.commit_media(obj, self.trans)
            self.update()
        return True


# ------------------------------------------------------------------------
# An op to convert relative paths to absolute
# ------------------------------------------------------------------------
class Convert2Abs(BatchOp):
    title = _("Convert paths from relative to _absolute")
    description = _(
        "This tool allows converting relative media paths "
        "to the absolute ones. It does this by prepending "
        "the base path as given in the Preferences, or if "
        "that is not set, it prepends user's directory."
    )

    def _prepare(self):
        self.set_total(self.db.get_number_of_media())
        with self.db.get_media_cursor() as cursor:
            for handle, data in cursor:
                obj = data_to_object(data)
                if not os.path.isabs(obj.path):
                    self.handle_list.append(handle)
                    self.path_list.append(obj.path)
                self.update()
        self.reset()

    def _run(self):
        if not self.prepared:
            self.prepare()
        self.set_total(len(self.handle_list))
        for handle in self.handle_list:
            obj = self.db.get_media_from_handle(handle)
            new_path = media_path_full(self.db, obj.path)
            obj.set_path(new_path)
            self.db.commit_media(obj, self.trans)
            self.update()
        return True


# ------------------------------------------------------------------------
# An op to convert absolute paths to relative
# ------------------------------------------------------------------------
class Convert2Rel(BatchOp):
    title = _("Convert paths from absolute to r_elative")
    description = _(
        "This tool allows converting absolute media paths "
        "to a relative path. The relative path is relative "
        "viz-a-viz the base path as given in the Preferences, "
        "or if that is not set, user's directory. "
        "A relative path allows to tie the file location to "
        "a base path that can change to your needs."
    )

    def _prepare(self):
        self.set_total(self.db.get_number_of_media())
        with self.db.get_media_cursor() as cursor:
            for handle, data in cursor:
                obj = data_to_object(data)
                if os.path.isabs(obj.path):
                    self.handle_list.append(handle)
                    self.path_list.append(obj.path)
                self.update()
        self.reset()

    def _run(self):
        if not self.prepared:
            self.prepare()
        self.set_total(len(self.handle_list))
        base_dir = media_path(self.db)
        for handle in self.handle_list:
            obj = self.db.get_media_from_handle(handle)
            new_path = relative_path(obj.path, base_dir)
            obj.set_path(new_path)
            self.db.commit_media(obj, self.trans)
            self.update()
        return True


# ------------------------------------------------------------------------
# An op to look for images that may have been forgotten.
# ------------------------------------------------------------------------
class ImagesNotIncluded(BatchOp):
    title = _("Add images not included in database")
    description = _("Check directories for images not included in database")
    description = _(
        "This tool adds images in directories that are "
        "referenced by existing images in the database."
    )

    def _prepare(self):
        """
        Get all of the fullpaths, and the directories of media
        objects in the database.
        """
        self.dir_list = set()
        self.set_total(self.db.get_number_of_media())
        with self.db.get_media_cursor() as cursor:
            for handle, data in cursor:
                obj = data_to_object(data)
                self.handle_list.append(handle)
                full_path = media_path_full(self.db, obj.path)
                self.path_list.append(full_path)
                directory, filename = os.path.split(full_path)
                if directory not in self.dir_list:
                    self.dir_list.add(directory)
                self.update()
        self.reset()

    def build_path_list(self):
        """
        This method returns a list of the path names that would be
        affected by the batch op. Typically it would rely on prepare()
        to do the actual job, but it does not have to be that way.
        """
        self.prepare()
        return self.dir_list

    def _run(self):
        """
        Go through directories that are mentioned in the database via
        media files, and include all images that are not all ready
        included.
        """
        if not self.prepared:
            self.prepare()
        self.set_total(len(self.dir_list))
        for directory in self.dir_list:
            for dirpath, dirnames, filenames in os.walk(directory):
                if ".git" in dirnames:
                    dirnames.remove(".git")  # don't visit .git directories
                for filename in filenames:
                    media_full_path = os.path.join(dirpath, filename)
                    if media_full_path not in self.path_list:
                        self.path_list.append(media_full_path)
                        mime_type = get_type(media_full_path)
                        if is_image_type(mime_type):
                            obj = Media()
                            obj.set_path(media_full_path)
                            obj.set_mime_type(mime_type)
                            (root, ext) = os.path.splitext(filename)
                            obj.set_description(root)
                            self.db.add_media(obj, self.trans)
            self.update()
        return True


# ------------------------------------------------------------------------
#
#
#
# ------------------------------------------------------------------------
class MediaManOptions(tool.ToolOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, person_id=None):
        tool.ToolOptions.__init__(self, name, person_id)
